#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Test case for an ingest path with files like those in
/lustre/aoc/cluster/pipeline/dsoc-dev/workspaces/staging/tmpx_ratuqh"""

import json
import tarfile
from pathlib import Path

import pytest
from conftest import (
    ADDITIONAL_METADATA_FILENAME,
    ANCILLARY_PRODUCTS,
    WEBLOG_ANCILLARY,
    find_ingestion_artifacts_tar,
    ingest_path,
    populate_fake_tmpx_ratuqh_ingest_path,
)
from ingest_envoy.ingestion_manifest import IngestionManifestBuilder
from ingest_envoy.manifest_components import (
    MANIFEST_FILENAME,
    TARFILE_EXT,
    WEBLOG_FILENAME,
    IngestionManifestKey,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import AncillaryProductType, ScienceProductType, Telescope

# pylint: disable=E0401, E0402, W0621


# pylint: disable=R1721


def test_image_manifest_creation(ingest_path: Path):
    """Test manifest creation for an image"""

    populate_fake_tmpx_ratuqh_ingest_path(ingest_path)
    manifest, manifest_file = build_tmpx_ratuqh_image_manifest(ingest_path)

    # Check that top-level groups match expectations
    mf_json = manifest.to_dict()
    keys = ["parameters", "input_group", "output_group"]
    assert len(mf_json) == len(keys)
    for key in keys:
        assert key in mf_json.keys()

    """
    The ancillary_products section of the manifest we build
    should match the one in the example.
    """
    output_group = manifest.output_group
    aps = output_group.ancillary_products
    assert len(aps) == len(ANCILLARY_PRODUCTS)

    # Check contents of ancillary_products
    weblog_candidates = [ap for ap in manifest.output_group.ancillary_products if ap.filename == WEBLOG_FILENAME]
    assert len(weblog_candidates) == 1
    assert weblog_candidates[0].type == AncillaryProductType.PIPELINE_WEBLOG

    ingest_artifacts_tar = find_ingestion_artifacts_tar(ingest_path)
    ingest_artifacts_candidates = [
        ap for ap in manifest.output_group.ancillary_products if ap.filename == ingest_artifacts_tar.name
    ]
    assert len(ingest_artifacts_candidates) == 1
    assert ingest_artifacts_candidates[0].type == AncillaryProductType.INGESTION_ARTIFACTS

    maybe_pips = [file for file in ingest_path.glob(AncillaryProductType.PIPELINE_ARTIFACTS.value + "*.tar")]
    assert len(maybe_pips) == 1
    pip_artie = maybe_pips[0]
    pipeline_artifacts_candidates = [
        ap for ap in manifest.output_group.ancillary_products if ap.filename == pip_artie.name
    ]
    assert len(pipeline_artifacts_candidates) == 1
    assert pipeline_artifacts_candidates[0].type == AncillaryProductType.PIPELINE_ARTIFACTS

    # Inspect the manifest's JSON dict
    """
    The output_group section of the manifest we build
    should match the one in the example:
    * a "science_products" section containing two science products comprising "type", "filename",
    and six ancillary products on each science product
    * an "ancillary products" section comprising three ancillary products
    """
    mf_json = manifest.to_dict()
    og_json = mf_json[IngestionManifestKey.OUTPUT_GROUP.value]
    assert len(og_json[IngestionManifestKey.SCIENCE_PRODUCTS.value]) == 2

    for sp_json in og_json[IngestionManifestKey.SCIENCE_PRODUCTS.value]:
        assert len(sp_json) == 3
        assert set(sp_json.keys()) == {IngestionManifestKey.ANCILLARY_PRODUCTS.value, "type", "filename"}
        # and ancillary products belonging to the first science product...
        sp_ap_jsons = sp_json[IngestionManifestKey.ANCILLARY_PRODUCTS.value]
        assert len(sp_ap_jsons) == 6

    # ... and ancillary products twisting in the wind all by themselves
    ap_jsons = og_json[IngestionManifestKey.ANCILLARY_PRODUCTS.value]
    assert len(ap_jsons) == 3

    # there should be a weblog, the ingestion artifacts tar, and the pipeline artifacts tar
    filenames_found = []
    for ap_json in ap_jsons:
        filename = ap_json["filename"]
        if filename.endswith(TARFILE_EXT):
            if filename.startswith(AncillaryProductType.INGESTION_ARTIFACTS.value) or filename.startswith(
                AncillaryProductType.PIPELINE_ARTIFACTS.value
            ):
                filenames_found.append(filename)
        elif filename == WEBLOG_FILENAME:
            filenames_found.append(filename)

    assert len(filenames_found) == 3

    """
    The ingestion_artifacts tar should contain ONLY
    the ingestion manifest and the additional metadata file
    """
    addl_md = manifest.parameters.additional_metadata
    ing_arties_tar = find_ingestion_artifacts_tar(ingest_path)
    with tarfile.open(ing_arties_tar, "r") as tar:
        members = tar.getmembers()
        assert len(members) == 2
        for member in members:
            assert member.name.endswith(MANIFEST_FILENAME) or member.name.endswith(addl_md.filename)


def test_catches_invalid_sp_type():
    """
    If we try to give IngestionManifestBuilder a science product type that doesn't translate into
    one of the ScienceProductType enum values, it should yell at us.

    :return:
    """

    additional_metadata = AbstractTextFile(filename="does-this-image-make-me-look-fat.json", content="")

    # we expect this to work fine
    IngestionManifestBuilder(
        manifest_source_dir=Path(),
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.IMAGE.value,
        locator="somewhere-over-the-rainbow",
        additional_metadata=additional_metadata,
    )
    # this, however, should fail
    with pytest.raises(ValueError):
        IngestionManifestBuilder(
            manifest_source_dir=Path(),
            telescope=Telescope.EVLA.value,
            sp_type="model_built_out_of_playdoh",
            locator="somewhere-over-the-rainbow",
            additional_metadata=additional_metadata,
        )


def test_catches_invalid_telescope():
    """
    If we try to give IngestionManifestBuilder a telescope that doesn't translate into
    one of the Telescope enum values, have no mercy.

    :return:
    """

    additional_metadata = AbstractTextFile(filename="arcane-knowledge.json", content="")

    # we expect this to work fine
    IngestionManifestBuilder(
        manifest_source_dir=Path(),
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.IMAGE.value,
        locator="ill-have-what-shes-having",
        additional_metadata=additional_metadata,
    )
    # this, however, should go belly-up
    with pytest.raises(ValueError):
        IngestionManifestBuilder(
            manifest_source_dir=Path(),
            telescope="nathans_mars_telescope",
            sp_type=ScienceProductType.IMAGE.value,
            locator="ill-have-what-shes-having",
            additional_metadata=additional_metadata,
        )


# -----------------------------
#  U  T  I  L  I  T  I  E  S
# -----------------------------


def build_tmpx_ratuqh_image_manifest(staging_source_dir: Path):
    """
    Build an image ingestion manifest that "should" match image_manifest_tmpx_ratuqh.json

    :return: the ingestion manifest and the manifest file in which it lives
    """

    additional_metadata = AbstractTextFile(filename=ADDITIONAL_METADATA_FILENAME, content="")

    return IngestionManifestBuilder(
        manifest_source_dir=staging_source_dir,
        additional_metadata=additional_metadata,
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.IMAGE.value,
        locator="uid://evla/calibration/ea93dae5-3495-47fa-887d-4be2852f5f14",
    ).build()
