"""
Facility for removing old or QAFailed files and directories from the Workspaces lustre areas
* Download area clean up is handled by the AAT/PPI mr_clean system
"""
import argparse
import logging
import os
import pathlib
import shutil
import sys
from argparse import ArgumentParser
from enum import Enum
from typing import Dict, List

import requests
from pycapo import CapoConfig
from ws_annihilator.connections import MDDBConnector

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


class Areas(Enum):
    CACHE = "cache"
    SPOOL = "spool"
    STAGING = "staging"
    STORAGE = "storage"


class TypedAnnihilators(Enum):
    """Sometimes we require cleanup times to be different by workflow type"""

    CURATOR = "curator"


def _get_settings(profile: str) -> Dict[str, str]:
    """
    Get annihilator capo settings

    :return:
    """
    config = CapoConfig(profile=profile)

    annihilator_settings = config.settings("edu.nrao.workspaces.AnnihilatorSettings")
    ingestion_settings = config.settings("edu.nrao.workspaces.IngestionSettings")

    workflow_url = config.settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl
    cache_directory = config.settings("edu.nrao.workspaces.DeliverySettings").cacheWeblogDirectory
    spool_directory = config.settings("edu.nrao.workspaces.ProcessingSettings").rootDirectory
    staging_directory = ingestion_settings.stagingDirectory
    storage_directory = ingestion_settings.storageDirectory

    return {
        "workflow_url": workflow_url,
        "cache": cache_directory,
        "spool": spool_directory,
        "storage": storage_directory,
        "staging": staging_directory,
        "cache_days": annihilator_settings.keepCacheForDays,
        "spool_days": annihilator_settings.keepSpoolForDays,
        "staging_days": annihilator_settings.keepStagingForDays,
        "storage_days": annihilator_settings.keepStorageForDays,
    }


def _arg_parser() -> ArgumentParser:
    """
    Create parser for annihilator

    :return:
    """
    arg_parser = ArgumentParser(
        description="Workspaces Directory Annihilator",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    arg_parser.add_argument(
        "-p",
        "--profile",
        action="store",
        required=True,
        help="Specify which environment to run on. Required. Example: dsoc-dev",
    )
    arg_parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        required=False,
        help="Run annihilation on workspaces spool, staging, and storage areas.",
    )
    arg_parser.add_argument(
        "-c",
        "--cache",
        action="store_true",
        required=False,
        help="Run annihilation on workspaces cache area.",
    )
    arg_parser.add_argument(
        "--curator",
        action="store_true",
        required=False,
        help="Run annihilation on completed curation directories only.",
    )
    return arg_parser


class Annihilator:
    """
    Class for removing workspaces directories that are no longer in use and past a given date

    - Retrieves a list of stale request directories to annihilate
    - Checks if directories still exist
        (DAs might manually clean up directories until scheduling system is available)
    - If directory exists in the workspaces' area for the specified profile, annihilates directory
        and sets 'cleaned' flag on the associated request
    - If directory doesn't exist, sets 'cleaned' flag on associated request
    """

    def __init__(self):
        self._args = _arg_parser().parse_args()
        self._settings = _get_settings(self._args.profile)

    def determine_path(self, area_to_clean: str) -> str:
        """
        return path to the location from capo settings

        :param area_to_clean: workspaces lustre area to clean, i.e. "spool"
        :return: path string
        """
        return self._settings.get(area_to_clean.lower())

    @staticmethod
    def path_valid(constructed_path: str, base_path: str) -> bool:
        """
        Check if the constructed path is valid for annihilation

        :param constructed_path: path to verify
        :param base_path: base path which directory should be annihilated from
        :return: True if path is valid for annihilation
        """

        return (
            constructed_path != base_path
            and constructed_path != base_path + "/"
            and len(constructed_path) > len(base_path)
        )

    def get_days_to_keep(self, area_to_clean: str) -> str:
        """
        Return the number of days to keep directories in a specified area from capo settings

        :param area_to_clean: workspaces lustre area to clean, i.e. "spool"
        :return:
        """
        key = area_to_clean.lower() + "_days"
        return self._settings.get(key)

    def get_stale_directories(self, area_to_clean: str) -> List[str]:
        """
        It is possible for processing to take multiple days to several weeks. While a job is running in HTCondor, it's
        parent workspaces directory in the spool area won't update other than the HTCondor log file(s).
        We want to be sure that only stale directories are annihilated.

        It is also possible for the dev and test systems to pick up production directories due to database copying.
        For this reason we only want to use the name of the directory itself and strip off the rest of the path.

        :param area_to_clean: workspaces lustre area to clean, i.e. "spool"
        :return: A list of directory names to be removed
        """

        annihilation_url = (
            self._settings.get("workflow_url") + "/workflows/requests/stale/" + self.get_days_to_keep(area_to_clean)
        )
        response = requests.get(annihilation_url)
        dir_list = response.json()["stale_dirs"]

        name_list = []
        # catch empty string response
        if dir_list:
            for stale in dir_list.split(","):
                # catch empty string directory or whitespace only directory or VLASS directory (VLASS will handle VLASS)
                if stale and not stale.isspace() and not stale.startswith("VLASS"):
                    name_list.append(pathlib.Path(str(stale)).name)

        return name_list

    def get_stale_directories_by_type(self, only_clean_type: str) -> List[str]:
        """
        Sometimes we need to clear out all directories meeting certain requirements before the standard clean date

        :param only_clean_type: type of directories to retrieve for cleaning
        :return: a list of directories to clean
        """
        name_list = []
        conn = MDDBConnector()

        try:
            cur = conn.cursor()

            sql = f"""
            SELECT results_dir 
            FROM workflow_requests
            WHERE workflow_name = %(wf_name)s AND state = 'Complete' AND cleaned = false
            """

            cur.execute(sql, {"wf_name": only_clean_type})
            data = cur.fetchall()
            if data:
                for wf_dir in data:
                    completed_wf = "".join(wf_dir)
                    print(completed_wf)
                    if completed_wf and not completed_wf.isspace() and self._args.profile in completed_wf:
                        name_list.append(pathlib.Path(completed_wf).name)
        except Exception as e:
            logger.error(f"ERROR: Unable to connect to metadata database. %s", exc_info=e)
        finally:
            conn.close()

        return name_list

    def annihilate_directory(self, area: str, directory: str):
        """
        Annihilate specified directory

        :param area: area being cleaned
        :param directory: directory to remove
        :return:
        """

        path = self.determine_path(area)
        clean_me = path + "/" + directory

        try:
            shutil.rmtree(directory)
        except PermissionError as pe:
            logger.error(
                f"ERROR: Permission Denied on directory {clean_me}. "
                f"Please check directory for non-vlapipe owned additions. %s",
                exc_info=pe,
            )
        except Exception as e:
            logger.error(f"ERROR: {clean_me} encountered the following problem: %s", exc_info=e)

        # After annihilation set cleaned flag...
        if area == Areas.SPOOL.value:
            logger.info(f"Annihilation complete for {directory}, setting 'cleaned' flag on workflow request...")
            self.set_cleaned(clean_me)
        else:
            logger.info(f"Annihilation complete for {directory} in {area}.")

    def set_cleaned(self, directory: str):
        """
        Set the 'cleaned' flag on the request of the specified directory

        :param directory: the directory that was annihilated
        :return:
        """
        clean_url = self._settings.get("workflow_url") + "/workflows/requests/cleaned"
        json_payload = {"directory": directory, "update_to": "True"}

        return requests.put(clean_url, json=json_payload)

    def annihilate(self, area_to_clean: str, only_clean_type: str = None):
        """
        Remove stale directories from targeted workspaces areas

        :param area_to_clean: what area on lustre are we cleaning? (spool, staging, etc.)
        :param only_clean_type: do we only want to annihilate a specific type of directory?
        :return:
        """
        if only_clean_type is not None:
            # get cleanable directories by wf type
            stale = self.get_stale_directories_by_type(only_clean_type)
        else:
            stale = self.get_stale_directories(area_to_clean)

        # If stale directories were found, annihilate them
        if stale:
            path = self.determine_path(area_to_clean)

            # change to area to clean
            os.chdir(path)

            for directory in stale:
                dir_path = path + "/" + directory
                if self.path_valid(dir_path, path):
                    if not pathlib.Path(dir_path).exists():
                        if area_to_clean == Areas.SPOOL.value:
                            logger.info(
                                f"directory {dir_path} not found, setting 'cleaned' flag on workflow request..."
                            )
                            self.set_cleaned(dir_path)
                        else:
                            logger.info(f"directory {dir_path} not found in {area_to_clean}, skipping...")
                    else:
                        logger.info(f"found directory {directory}, annihilating...")
                        self.annihilate_directory(area_to_clean, directory)
                else:
                    logger.error(f"Error: Received invalid directory '{dir_path}' as input for annihilation. Aborting.")
                    exit(1)

    def run(self):
        """
        run the annihilator based on provided args

        :return:
        """
        if self._args.all:
            logger.info(f"Starting cleaning of Workspaces Spool for profile {self._args.profile}")
            self.annihilate(Areas.SPOOL.value)

            logger.info(f"Starting cleaning of Workspaces Staging for profile {self._args.profile}")
            self.annihilate(Areas.STAGING.value)

            logger.info(f"Starting cleaning of Workspaces Storage for profile {self._args.profile}")
            self.annihilate(Areas.STORAGE.value)
            logger.info(
                f"Finished cleaning Workspaces Spool, Staging, and Storage areas for profile {self._args.profile}!"
            )
        elif self._args.cache:
            logger.info(f"Starting cleaning of Workspaces Cache for profile {self._args.profile}")
            self.annihilate(Areas.CACHE.value)
            logger.info(f"Finished cleaning Workspaces Cache for profile {self._args.profile}!")
        elif self._args.curator:
            logger.info(f"Starting cleaning of completed Curator workflows for profile {self._args.profile}")
            self.annihilate(Areas.SPOOL.value, TypedAnnihilators.CURATOR.value)
        else:
            logger.info(f"Starting cleaning of Workspaces Spool for profile {self._args.profile}")
            self.annihilate(Areas.SPOOL.value)
            logger.info(f"Finished cleaning Workspaces Spool for profile {self._args.profile}!")


def main():
    Annihilator().run()
