#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any, Dict, Iterable, List, Optional

from .interfaces import Relationship, RowSet, RowWriter, Table


class PGTable(Table):
    """
    I'm a table in PostgreSQL.
    """

    @property
    def name(self) -> str:
        return self._name

    def __init__(
        self,
        cursor,
        name: str,
    ):
        self.cursor = cursor
        self._name = name

    def fetch(self, primary_key: Dict[str, Any]) -> RowSet:
        # 1. Determine the primary key columns
        primary_key_columns = self.primary_key_columns()

        # 2. Generate the WHERE clause
        pkey = {column_name: primary_key[column_name] for column_name in primary_key_columns}
        whereclause = " AND ".join(f"{name} = %({name})s" for name in pkey.keys())

        # 3. Run the query
        self.cursor.execute(f"SELECT * FROM {self.name} WHERE {whereclause}", pkey)

        # 4. Manufacture the result
        return PGRowSet(self, self.cursor.fetchall())

    def fetch_from_many(self, primary_keys: Dict[str, Any]) -> RowSet:
        primary_key_columns = self.primary_key_columns()
        pkeys = {column_name: primary_keys[column_name] for column_name in primary_key_columns}
        whereclause = " AND ".join(f"{name} = ANY(%({name})s)" for name in pkeys.keys())
        self.cursor.execute(f"SELECT * FROM {self.name} WHERE {whereclause}", pkeys)
        return PGRowSet(self, self.cursor.fetchall())

    def primary_key_columns(self):
        self.cursor.execute(
            """SELECT c.column_name
                 FROM information_schema.table_constraints tc
                 JOIN information_schema.constraint_column_usage AS ccu USING (constraint_schema, constraint_name)
                 JOIN information_schema.columns AS c ON c.table_schema = tc.constraint_schema AND
                                                         tc.table_name = c.table_name AND
                                                         ccu.column_name = c.column_name
                WHERE constraint_type = 'PRIMARY KEY' AND
                      tc.table_name = %(tablename)s""",
            dict(tablename=self.name),
        )
        primary_key_columns = [r["column_name"] for r in self.cursor.fetchall()]
        return primary_key_columns

    def fetch_according_to(self, rows: RowSet, columns: List[str]):
        self_join_columns = self.self_join_columns()
        if self_join_columns:
            return self.recursive_fetch_according_to(rows, columns, self_join_columns)
        else:
            return self.basic_fetch_according_to(rows, columns)

    def basic_fetch_according_to(self, rows: RowSet, columns: List[str]):
        """
        Fetch from this table all the rows related to the supplied rowset, using the
        supplied columns from the originating table.

        :param rows:
        :param columns:
        :return:
        """
        query = self.basic_fetch_according_to_query(rows, columns)

        # if the query doesn't exist, we don't have any escaped rows,
        # so there's nothing to do
        if query is None:
            return PGRowSet(self, [])

        self.cursor.execute(query)
        return PGRowSet(self, self.cursor.fetchall())

    def basic_fetch_according_to_query(self, rows: RowSet, columns: List[str]) -> Optional[str]:
        """
        Returns the query we'll use to obtain some rows from this table according to rows in
        another RowSet that reference us somehow.

        :param rows:     the rows with references to our table
        :param columns:  the columns in those rows that reference our primary key
        :return: SQL query string
        """
        # 1. Escape the WHERE clause entries. it's important to use the primary keys
        #    to retrieve the values from the previous resultset; the new columns will
        #    appear in the query below.
        key_columns = rows.table.primary_key_columns()
        escaped = [self.escape(row[column]) for column in key_columns for row in rows if row[column] is not None]

        # If we have no escaped entries, then there won't be anything in the resultset and we can return now
        if len(escaped) == 0:
            return None

        # 2. Build the WHERE clause
        whereclause = f"({','.join(columns)}) IN ({','.join(escaped)})"

        # 3. Send the query off and make the result
        return f"SELECT * FROM {self.name} WHERE {whereclause}"

    def recursive_fetch_according_to(self, rows: RowSet, columns: List[str], self_columns: List[str]):
        """
        For tables that have self-joins, we have to make a slightly more complex query. In
        fact we need a recursive query that starts from the rows referencing us but encompasses
        all of the children of those rows. To do that, we'll need to know the self-join columns.

        :param rows:          rows in the originating table
        :param columns:       the columns that reference our primary key in those rows
        :param self_columns:  the self-join columns we have
        :return:
        """
        basic_select = self.basic_fetch_according_to_query(rows, columns)

        if basic_select is None:
            return PGRowSet(self, [])

        on_clause = " AND ".join(
            f"{self.name}.{fk} = recursive_{self.name}.{pk}" for pk, fk in zip(self.primary_key_columns(), self_columns)
        )

        query = f"""with recursive recursive_{self.name} as (
            {basic_select}
            union
            select {self.name}.* from {self.name}
            join recursive_{self.name} on {on_clause}
            )
        select * from recursive_{self.name}"""

        self.cursor.execute(query)
        return PGRowSet(self, self.cursor.fetchall())

    @staticmethod
    def escape(value: Any):
        """
        Escape the supplied value so that Postgres can understand it
        :param value: some value, a string or something
        :return: an escaped value
        """
        if isinstance(value, str):
            return f"E{repr(value)}"
        elif value is None:
            return None
        else:
            return str(value)

    def relations(self) -> Iterable[Relationship]:
        # To obtain the other relationships, we must first retrieve the foreign keys, and then
        # convert them into objects in this schema.
        self.cursor.execute(
            self.foreign_keys_query(self_only=False),
            dict(tablename=self.name),
        )

        # now that we have the rows, we can construct our return value
        result = []
        for row in self.cursor.fetchall():
            other_table = PGTable(self.cursor, row["foreign_table"])
            result.append(PGRelationship(self, other_table, row["fk_columns"].split(",")))

        return result

    def self_join_columns(self) -> List[str]:
        """
        Determine if there are any self-join columns on this table, and if so, return them.
        :return: A list of self-join columns, possibly empty.
        """
        self.cursor.execute(
            self.foreign_keys_query(self_only=True),
            dict(tablename=self.name),
        )
        rows = self.cursor.fetchall()
        if len(rows) > 0:
            return rows[0]["fk_columns"].split(",")
        else:
            return []

    @staticmethod
    def foreign_keys_query(self_only=False) -> str:
        """
        Return a query we can use to find the foreign keys related to this table.

        :param self_only: if True, only self-joins. if False, only other-joins.
        :return: A string query we can use.
        """

        # PostgreSQL has extensive metadata about the database stored in the information_schema, which
        # is an SQL standard suite of views. In Postgres, these views are based on lower-level Postgres
        # metadata in the pg_catalog schema.
        #
        # This precise query gets you the foreign keys from this table to other tables. It explicitly
        # omits self-joins, because they're handled in this class when you retrieve.

        return f"""select kcu.table_name                    as foreign_table,
                       string_agg(kcu.column_name, ',') as fk_columns
                from information_schema.table_constraints tco
                join information_schema.key_column_usage kcu
                     on tco.constraint_schema = kcu.constraint_schema
                         and tco.constraint_name = kcu.constraint_name
                join information_schema.referential_constraints rco
                     on tco.constraint_schema = rco.constraint_schema
                         and tco.constraint_name = rco.constraint_name
                join information_schema.table_constraints rel_tco
                     on rco.unique_constraint_schema = rel_tco.constraint_schema
                         and rco.unique_constraint_name = rel_tco.constraint_name
                where tco.constraint_type = 'FOREIGN KEY'
                  and rel_tco.table_name = %(tablename)s
                                  and kcu.table_name {'=' if self_only else '<>'} %(tablename)s
                group by kcu.table_name,
                    rel_tco.table_name,
                    rel_tco.table_schema,
                    kcu.constraint_name
                order by kcu.table_name"""

    def delete(self, rows: RowSet):
        for row in rows:
            whereclause = "".join(
                            f" AND {column_name} = '{row[column_name]}'"
                            if row[column_name] is not None and row[column_name] != ''
                            else "" for column_name in row)
            whereclause = whereclause[5:]
            self.cursor.execute(f"DELETE FROM {self.name} WHERE {whereclause}")

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return f"<Table {self.name}>"


class PGRowSet(RowSet):
    """
    I'm a RowSet from a Postgres database.
    """

    def __init__(self, table: Table, rows: List[Dict]):
        self._table, self.rows = table, rows

    @property
    def table(self) -> Table:
        return self._table

    def relations(self) -> Iterable[Relationship]:
        """
        Return the relationships that might pertain to the rows we have here.

        :return:
        """
        return self.table.relations()

    def write_to(self, writer: RowWriter):
        writer.write_rows(self.table, self.rows)

    def delete(self):
        self.table.delete(self)

    def combine_rows(self, other):
        self.rows = self.rows + other.rows

    def __iter__(self) -> Iterable[Dict]:
        return iter(self.rows)

    def __len__(self) -> int:
        return len(self.rows)

    def __repr__(self):
        return f"<RowSet of {self.table} with {len(self.rows)} rows>"


class PGRelationship(Relationship):
    """
    I'm a relationship between Postgres tables.
    """

    def __init__(self, source_table: Table, destination_table: Table, foreign_columns: List[str]):
        self._source_table, self._destination_table = source_table, destination_table
        self.foreign_columns = foreign_columns

    @property
    def source_table(self) -> Table:
        return self._source_table

    @property
    def destination_table(self) -> Table:
        return self._destination_table

    def fetch_related_to(self, rows: RowSet) -> RowSet:
        """
        Return rows related to the supplied rows based on this relationship.
        For instance, if this is a relationship from projects to science products,
        you can supply a RowSet of projects to this relationship, and receive a
        new RowSet of science products.

        :param rows:  source rows
        :return:      rows from the destination
        """
        return self.destination_table.fetch_according_to(rows, self.foreign_columns)

    def __repr__(self):
        return f"<Relationship from {self.source_table} to {self.destination_table}>"
